#!/usr/bin/env python
from scipy.ndimage import gaussian_filter
from sklearn.decomposition import PCA
import pandas as pd
import random, itertools
import matplotlib.colors as mcolors
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import os
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas



def calculate_orient(matrix, bp1, bp2, bpLink, chrom1, chrom2):
    '''
    Calculate the orient
    '''
    # Location peak
    index = np.unravel_index(np.argmax(matrix), np.array(matrix).shape)
    cur_row = np.shape(matrix)[0] // 2
    cur_col = np.shape(matrix)[1] // 2
    if 0 <= index[0] < cur_row and 0 <= index[1] < cur_col:
        row_peak = -1
        col_peak = 1
    elif 0 <= index[0] < cur_row and index[1] >= cur_col:
        row_peak = -1
        col_peak = -1
    elif index[0] >= cur_row and 0 <= index[1] < cur_col:
        row_peak = 1
        col_peak = 1
    else:
        row_peak = 1
        col_peak = -1
    # row and col correlation
    row_pos = np.arange(1, np.shape(matrix)[0] + 1)
    col_pos = np.arange(1, np.shape(matrix)[1] + 1)
    cols = np.mean(matrix, axis=0)
    rows = np.mean(matrix, axis=1)
    rows_coff = np.corrcoef(row_pos, rows)[0, 1]
    cols_coff = -np.corrcoef(col_pos, cols)[0, 1]
    # diagonal correlation
    d = np.diag(matrix)
    diag_rows_coff = np.corrcoef(rows, d)[0, 1]
    diag_cols_coff = np.corrcoef(cols, d)[0, 1]
    sum_row = row_peak + rows_coff + diag_rows_coff
    sum_col = col_peak + cols_coff + diag_cols_coff
    if sum_row > 0 and sum_col > 0:
        bpLink.loc[len(bpLink.index)] = [chrom1, bp1, '+', chrom2, bp2, '+']
    elif sum_row > 0 > sum_col:
        bpLink.loc[len(bpLink.index)] = [chrom1, bp1, '+', chrom2, bp2, '-']
    elif sum_row < 0 < sum_col:
        bpLink.loc[len(bpLink.index)] = [chrom1, bp1, '-', chrom2, bp2, '+']
    else:
        bpLink.loc[len(bpLink.index)] = [chrom1, bp1, '-', chrom2, bp2, '-']
    return bpLink


def calculate_std(matrix):
    count = []
    row, col = np.shape(matrix)
    for i in range(row):
        for j in range(col):
            if matrix[i][j] != 0:
                count.append(matrix[i][j])
    # remove outlier
    count = np.array(count)
    q1 = np.quantile(count, 0.25)
    q3 = np.quantile(count, 0.75)
    iqr = q3 - q1
    upper_bound = q3 + (3 * iqr)
    index = count < upper_bound

    return np.std(count[index])


def read_SV_info(bp_file):
    '''
        read SV breakpoint information
        :param breakpoint file name :
        :return: {'chrom1', 'pos1', 'chrom2', 'pos2'}
    '''
    result = pd.DataFrame(columns=['chrom1', 'pos1', 'chrom2', 'pos2'])
    with open(bp_file) as f:
        for line in f:
            line_data = line.strip().split('\t')
            result.loc[len(result.index)] = [line_data[0], int(line_data[1]), line_data[2], int(line_data[3])]
    return result


def extend_stretches(arr, min_seed_len=2, max_gap=1, min_stripe_len=5):
    arr.sort()
    pieces = np.split(arr, np.where(np.diff(arr) != 1)[0] + 1)
    filtered = [p for p in pieces if len(p) >= min_seed_len]  # can't be empty
    stripes = []
    seed = filtered[0]
    for p in filtered[1:]:
        if p[0] - seed[-1] < (max_gap + 2):
            seed = np.r_[seed, p]
        else:
            if seed[-1] - seed[0] + 1 >= min_stripe_len:
                stripes.append([seed[0], seed[-1] + 1])
            seed = p

    if seed[-1] - seed[0] + 1 >= min_stripe_len:
        stripes.append([seed[0], seed[-1] + 1])
    return stripes


def locate(pca1, left_most=True):

    loci_1 = extend_stretches(np.where(pca1 >= 0)[0])
    loci_2 = extend_stretches(np.where(pca1 <= 0)[0])
    loci = loci_1 + loci_2
    loci.sort()
    if left_most:
        b = loci[0][1] - 1
    else:
        b = loci[-1][0]

    return b


def recalculate_breakpoint(matrix, orient1, orient2):
    '''

    :param SV_region matrix:
    :return: breakpoint position
    '''
    rowmask = matrix.sum(axis=1) != 0
    colmask = matrix.sum(axis=0) != 0
    new = matrix[rowmask][:, colmask]
    # row
    corr = gaussian_filter(np.corrcoef(new, rowvar=True), sigma=2)
    corr[np.isnan(corr)] = 0
    try:
        pca = PCA(n_components=3, whiten=True)
        pca1 = pca.fit_transform(corr)[:, 0]
        if orient1 == '+':
            loc = locate(pca1, left_most=False)
        else:
            loc = locate(pca1, left_most=True)
        up_i = np.where(rowmask)[0][loc]  # included
    except:
        up_i = 0
    # col
    corr = gaussian_filter(np.corrcoef(new, rowvar=False), sigma=2)
    try:
        pca = PCA(n_components=3, whiten=True)
        pca1 = pca.fit_transform(corr)[:, 0]
        if orient2 == '+':
            loc = locate(pca1, left_most=True)
        else:
            loc = locate(pca1, left_most=False)
        down_i = np.where(colmask)[0][loc]
    except:
        down_i = 0
    return up_i, down_i


def get_MatrixFile(hic_path, name):
    matrix = 0
    for filewalks in os.walk(hic_path):
        for files in filewalks[2]:
            if name in files:
                matrix = np.loadtxt(os.path.join(filewalks[0], files))
    return matrix


def get_SV_orient(bp_result, dict_matrix, res, merge):
    # init orient
    bpLink = pd.DataFrame(columns=['chrom1', 'pos1', 'orient1', 'chrom2', 'pos2', 'orient2'])
    overlap = False
    for k in bp_result.index:
        row = bp_result.loc[k]
        cur_chr1 = row['chrom1']
        cur_chr2 = row['chrom2']
        name = str(cur_chr1) + '_' + str(cur_chr2)
        matrix = dict_matrix[name]
        pos1 = row['pos1']
        pos2 = row['pos2']
        if pos1 < merge or pos2 < merge or (pos1 + merge) // res > np.shape(matrix)[0] or (pos2 + merge) // res > np.shape(matrix)[1]:
            merge = min(pos1, pos2, np.shape(matrix)[0]*res-pos1,  np.shape(matrix)[1]*res-pos2)
        pos1_start = (pos1 - merge) // res
        pos1_end = (pos1 + merge) // res
        pos2_start = (pos2 - merge) // res
        pos2_end = (pos2 + merge) // res
        cur_matrix = matrix[pos1_start:pos1_end, pos2_start:pos2_end]
        # instead of zero bin
        for m in range(np.shape(cur_matrix)[0]):
            for n in range(np.shape(cur_matrix)[1]):
                if cur_matrix[m][n] == 0:
                        cur_matrix[m][n] = round(random.uniform(0, 1), 2)

        row, col = np.shape(cur_matrix)
        # calculate std cut-off value
        cut_off = calculate_std(cur_matrix)
        cut_off *= 2
        # calculate orient
        cut_1 = merge // res
        cut_2 = merge // res
        cut_matrix = {}
        cut_std = []
        cur_cut_matrix = cur_matrix[0:cut_1, 0:cut_2]
        cut_std.append(calculate_std(cur_cut_matrix))
        cut_matrix[0] = cur_cut_matrix
        cur_cut_matrix = cur_matrix[cut_1:row, 0:cut_2]
        cut_std.append(calculate_std(cur_cut_matrix))
        cut_matrix[1] = cur_cut_matrix
        cur_cut_matrix = cur_matrix[0:cut_1, cut_2:col]
        cut_std.append(calculate_std(cur_cut_matrix))
        cut_matrix[2] = cur_cut_matrix
        cur_cut_matrix = cur_matrix[cut_1:row, cut_2:col]
        cut_std.append(calculate_std(cur_cut_matrix))
        cut_matrix[3] = cur_cut_matrix
        count = sum(i > cut_off for i in cut_std)
        if count == 2:
            filtered = filter(lambda x: x > cut_off, cut_std)
            pos = list(map(cut_std.index, filtered))
            if pos[0] == 0 and pos[1] == 3:
                bpLink = calculate_orient(cut_matrix[0], pos1 // res, pos2 // res, bpLink, cur_chr1, cur_chr2)
                bpLink = calculate_orient(cut_matrix[3], pos1 // res, pos2 // res, bpLink, cur_chr1, cur_chr2)
            elif pos[0] == 1 and pos[1] == 2:
                bpLink = calculate_orient(cut_matrix[1], pos1 // res, pos2 // res, bpLink, cur_chr1, cur_chr2)
                bpLink = calculate_orient(cut_matrix[2], pos1 // res, pos2 // res, bpLink, cur_chr1, cur_chr2)
            else:
                overlap = True
                index = cut_std.index(max(cut_std))
                bpLink = calculate_orient(cut_matrix[index], pos1 // res, pos2 // res, bpLink, cur_chr1, cur_chr2)

        else:
            index = cut_std.index(max(cut_std))
            bpLink = calculate_orient(cut_matrix[index], pos1 // res, pos2 // res, bpLink, cur_chr1, cur_chr2)

    return bpLink, overlap


def fragment_boundary(chrom, start, end, orient, index, SV_fragment):
    cur_start = start
    cur_end = end
    num = [i for i in range(SV_fragment.shape[0])]
    del num[index]
    for i in num:
        row = SV_fragment.loc[i]
        if row['chrom1'] == chrom:
            fragment1 = [m for m in range(start, end+1)]
            fragment2 = [n for n in range(row['start1'], row['end1']+1)]
            ret1 = list(set(fragment1).intersection(set(fragment2)))
            if len(ret1) > (len(fragment1) * 0.6) and len(ret1) > (len(fragment2) * 0.6):
                if orient == '+' and row['orient1'] == '-':
                    cur_start = start
                    cur_end = row['end1']
                elif orient == '+' and row['orient1'] == '+':
                    cur_start = min(row['start1'], start)
                    cur_end = max(row['end1'], end)
                elif orient == '-' and row['orient1'] == '-':
                    cur_start = min(row['start1'], start)
                    cur_end = max(row['end1'], end)
                else:
                    cur_start = row['start1']
                    cur_end = end

        if row['chrom2'] == chrom:
            fragment1 = [m for m in range(start, end + 1)]
            fragment2 = [n for n in range(row['start2'], row['end2'] + 1)]
            ret1 = list(set(fragment1).intersection(set(fragment2)))
            if len(ret1) > (len(fragment1) * 0.6) and len(ret1) > (len(fragment2) * 0.6):
                # print(chrom, start, end)
                # print(row['chrom1'], row['start1'], row['end1'] + 1)
                if orient == '+' and row['orient2'] == '-':
                    cur_start = start
                    cur_end = row['end2']
                elif orient == '+' and row['orient2'] == '+':
                    cur_start = min(row['start2'], start)
                    cur_end = max(row['end2'], end)
                elif orient == '-' and row['orient2'] == '-':
                    cur_start = min(row['start2'], start)
                    cur_end = max(row['end2'], end)
                else:
                    cur_start = row['start2']
                    cur_end = end

    return cur_start, cur_end


def coverFragment(SV_fragment):
    for i in SV_fragment.index:
        row = SV_fragment.loc[i]
        start1, end1 = fragment_boundary(row['chrom1'], row['start1'], row['end1'], row['orient1'], i, SV_fragment)
        start2, end2 = fragment_boundary(row['chrom2'], row['start2'], row['end2'], row['orient2'], i, SV_fragment)
        SV_fragment.loc[i, 'start1'] = start1
        SV_fragment.loc[i, 'end1'] = end1
        SV_fragment.loc[i, 'start2'] = start2
        SV_fragment.loc[i, 'end2'] = end2
    return SV_fragment


def genomeFragment(SV_fragment):
    chrom = []
    start = []
    end = []
    node_list = []
    for i in SV_fragment.index:
        row = SV_fragment.loc[i]
        start1 = row['start1']
        end1 = row['end1']
        start2 = row['start2']
        end2 = row['end2']
        if start1 not in start and end1 not in end:
            start.append(start1)
            end.append(end1)
            chrom.append(row['chrom1'])
            node_list.append(row['fragment1'])
        if start2 not in start and end2 not in end:
            start.append(start2)
            end.append(end2)
            chrom.append(row['chrom2'])
            node_list.append(row['fragment2'])
        if start1 in start and end1 in end:
            index = start.index(start1)
            if chrom[index] != row['chrom1']:
                start.append(start1)
                end.append(end1)
                chrom.append(row['chrom1'])
                node_list.append(row['fragment1'])
        if start2 in start and end2 in end:
            index = start.index(start2)
            if chrom[index] != row['chrom2']:
                start.append(start2)
                end.append(end2)
                chrom.append(row['chrom2'])
                node_list.append(row['fragment2'])

    return chrom, start, end, node_list

segment_to_node = {}
node_labels = itertools.chain(
    (chr(i) for i in range(65, 91)),  # Single letters A-Z
    (''.join(pair) for pair in itertools.product("ABCDEFGHIJKLMNOPQRSTUVWXYZ", repeat=2))  # Double letters AA, AB, etc.
)
def get_node_name(chrom, start, end):
    segment_key = (chrom, start, end)
    if segment_key not in segment_to_node:
        segment_to_node[segment_key] = next(node_labels)
    return segment_to_node[segment_key]

def constructPath(SV_fragment):
    all_path = []
    all_orient = []
    SV_fragment['flag'] = 0
    SV_fragment.sort_values(["chrom1", "start1"], inplace=True)
    SV_fragment.reset_index(drop=True, inplace=True)
    SV_fragment['fragment1'] = SV_fragment.apply(lambda row: get_node_name(row['chrom1'], row['start1'], row['end1']), axis=1)
    SV_fragment['fragment2'] = SV_fragment.apply(lambda row: get_node_name(row['chrom2'], row['start2'], row['end2']), axis=1)
    # path
    while SV_fragment['flag'].sum() != SV_fragment.shape[0]:
        index = SV_fragment[SV_fragment['flag'] == 0].index[0]
        init_link = SV_fragment.loc[index]
        node = init_link['fragment1']
        next_orient = init_link['orient2']
        next_node = init_link['fragment2']
        path_list = []
        orient_list = []
        path_list.append(node)
        path_list.append(next_node)
        orient_list.append(init_link['orient1'])
        orient_list.append(next_orient)
        SV_fragment.loc[index, 'flag'] = 1
        count = 1
        while count < SV_fragment.shape[0]:
            row = SV_fragment.loc[count]
            if row['flag'] == 0 and row['orient1'] == next_orient and next_node == row['fragment1']:
                next_node = row['fragment2']
                next_orient = row['orient2']
                path_list.append(next_node)
                orient_list.append(next_orient)
                SV_fragment.loc[count, 'flag'] = 1
                count = 1
            elif row['flag'] == 0 and row['orient2'] != next_orient and next_node == row['fragment2'] :
                next_node = row['fragment1']
                if row['orient1'] == '+':
                    next_orient = '-'
                else:
                    next_orient = '+'
                path_list.append(next_node)
                orient_list.append(next_orient)
                SV_fragment.loc[count, 'flag'] = 1
                count = 1
            else:
                count += 1
        all_path.append(path_list)
        all_orient.append(orient_list)
    return all_path, all_orient


def plot_SV_genome(path, all_orient, sv_fragment):
    fig = plt.figure(figsize=(10, 4))
    ax = fig.add_subplot(111)
    h = 0
    length_list = []
    CGR_info = {'num': [], 'chrom': [], 'start': [], 'end': [], 'node': [], 'orient': []}
    for i in range(len(path)):
        cur_start = 0
        count = 0
        colors = list(mcolors.TABLEAU_COLORS.keys())
        for j in range(len(path[i])):
            first_match = sv_fragment[(sv_fragment['fragment1'] == path[i][j]) | (sv_fragment['fragment2'] == path[i][j])].iloc[0]
            if first_match['fragment1'] == path[i][j]:
                fragment_chrom = first_match['chrom1']
                fragment_start = first_match['start1']
                fragment_end = first_match['end1']
            if first_match['fragment2'] == path[i][j]:
                fragment_chrom = first_match['chrom2']
                fragment_start = first_match['start2']
                fragment_end = first_match['end2']
            CGR_info['num'].append(i)
            CGR_info['chrom'].append(fragment_chrom)
            CGR_info['start'].append(fragment_start)
            CGR_info['end'].append(fragment_end)
            CGR_info['node'].append(path[i][j])
            CGR_info['orient'].append(all_orient[i][j])
            length = fragment_end - fragment_start + 1
            count += length
            textPos = cur_start + length // 2
            plt.text(textPos, h+2, path[i][j], fontsize=10, weight='bold')
            if all_orient[i][j] == '+':
                pos1 = cur_start
                pos2 = pos1 + length
            else:
                pos2 = cur_start
                pos1 = pos2 + length
            ax.annotate("",
                        xy=[pos2, h],
                        xytext=[pos1, h],
                        size=4,
                        arrowprops=dict(color=mcolors.TABLEAU_COLORS[colors[i % len(colors)]], headwidth=8, width=4, headlength=4), zorder=0)

            cur_start = cur_start + length + 1
        length_list.append(count)
        h -= 10

    plt.xlim([0, max(length_list)+5])
    plt.ylim([h, 3])
    plt.axis('off')
    canvas = FigureCanvas(fig)
    return canvas, CGR_info



def combine_matrix(dict_matrix, dict_start, dict_end, chrom_list, resolution):
    # init_matrix
    sum_bin = 0
    for i in range(len(chrom_list)):
        sum_bin += (dict_end[chrom_list[i]] - dict_start[chrom_list[i]])
    all_matrix = np.full((sum_bin, sum_bin), 0)
    chrom_1 = 0
    interval = []
    text_pos = []
    text = []
    for i in range(len(chrom_list)):
        start1 = dict_start[chrom_list[i]]
        end1 = dict_end[chrom_list[i]]
        length1 = end1 - start1
        chrom_2 = chrom_1
        interval.append(chrom_1)
        text_pos.append(chrom_1)
        text_pos.append(chrom_1+length1)
        text.append(str(start1*resolution // 1000000) + 'M')
        text.append(str(end1*resolution // 1000000) + 'M')
        for j in range(i, len(chrom_list)):
            name = str(chrom_list[i]) + '_' + str(chrom_list[j])
            start2 = dict_start[chrom_list[j]]
            end2 = dict_end[chrom_list[j]]
            length2 = end2 - start2
            all_matrix[chrom_1:chrom_1+length1, chrom_2:chrom_2+length2] = dict_matrix[name][start1:end1, start2:end2]
            chrom_2 += length2
        chrom_1 += length1
    return all_matrix, interval, text_pos, text


def matrix_plot(matrix, interval, chrom_text_pos, chrom_text,
                SV_fragment, chrom, start, end, node_list, chrom_list, plot_start, plot_end):
    
    fig, (ax_heatmap, ax_genome) = plt.subplots(2, 1, 
                                                gridspec_kw={'height_ratios': [8, 3]}, figsize=(10, 10), sharex=True)
    fig.subplots_adjust(hspace=0.05) 
    cdict = {
        'red': ((0.0, 0.0, 1.0), (1.0, 1.0, 1.0)),
        'green': ((0.0, 0.0, 1.0), (1.0, 0.0, 0.0)),
        'blue': ((0.0, 0.0, 1.0), (1.0, 0.0, 0.0))
    }

    cmap = LinearSegmentedColormap('Rd_Bl_Rd', cdict, 256)
    M = matrix
    n = M.shape[0]
    vmax = np.percentile(M, 96)
    vmin = M.min()
    # Create the rotation matrix
    t = np.array([[1, 0.5], [-1, 0.5]])
    A = np.dot(np.array([(i[1], i[0]) for i in itertools.product(range(n, -1, -1), range(0, n + 1, 1))]), t)
    # Plot the Heatmap ...
    x = A[:, 1].reshape(n + 1, n + 1)
    y = A[:, 0].reshape(n + 1, n + 1)
    y[y < 0] = -y[y < 0]
    ax_heatmap.pcolormesh(x, y, np.flipud(M), vmin=vmin, vmax=vmax, cmap=cmap,
                  edgecolor='none', snap=True, linewidth=.01, rasterized=True)
    # plot boundary
    for i in range(len(interval)):
        if interval[i] != 0:
            ax_heatmap.plot([interval[i], x[n, interval[i]]], [0, y[n, interval[i]]],
                     color="black", linestyle='--', linewidth=2)
            ax_heatmap.plot([interval[i], x[(n-interval[i]), n]], [0, y[(n-interval[i]), n]],
                     color="black", linestyle='--', linewidth=2)
            
    ax_heatmap.plot([0, x[0, 0]], [0, y[0, 0]], color="black", linewidth=2)
    ax_heatmap.plot([x[0, 0], n], [y[0, 0], 0], color="black", linewidth=2)
    ax_heatmap.axis('off')
    ## plot genome
    color_list = []
    chrom_start = 0
    dict_start = {}
    for i in range(len(chrom_list)):
        count = 0
        dict_start[chrom_list[i]] = chrom_start
        chrom_length = plot_end[chrom_list[i]]-plot_start[chrom_list[i]]
        ax_genome.add_patch(patches.Rectangle((chrom_start, 0), width=chrom_length, height=0.8, color='gray', zorder=1, alpha=.4, linewidth=1, edgecolor='black'))
        rect_mid = chrom_start + chrom_length / 2
        ax_genome.text(rect_mid, 0.8, chrom_list[i], ha='center', va='center', fontsize=10, color='black')
        ax_genome.text(chrom_start, 0.8, f"{chrom_text[count]}", ha='right', va='center', fontsize=8, color='black')
        ax_genome.text(chrom_start + chrom_length, 0.8, f"{chrom_text[count+1]}", ha='left', va='center', fontsize=8, color='black')
        chrom_start += chrom_length
        count += 1
    ## plot fragment
    for i in range(len(start)):
        cur_start = dict_start[chrom[i]]
        rect_start = cur_start + (start[i] - plot_start[chrom[i]])
        rect_width = end[i] - start[i]
        ax_genome.add_patch(patches.Rectangle((rect_start, 0), width=rect_width, height=0.8, color='black', zorder=1, alpha=.4, linewidth=0))
        rect_mid = rect_start + rect_width / 2
        ax_genome.text(rect_mid, 0.3, node_list[i], ha='center', va='center', fontsize=10, color='black')
    ax_genome.axis('off')
    ## plot link
    h = -0.5
    for i in SV_fragment.index:
        row = SV_fragment.loc[i]
        start1 = (row['start1']-plot_start[row['chrom1']]) + dict_start[row['chrom1']]
        end1 = (row['end1']-plot_start[row['chrom1']]) + dict_start[row['chrom1']]
        start2 = (row['start2']-plot_start[row['chrom2']]) + dict_start[row['chrom2']]
        end2 = (row['end2']-plot_start[row['chrom2']]) + dict_start[row['chrom2']]
        colors = list(mcolors.TABLEAU_COLORS.keys())
        color_list.append(mcolors.TABLEAU_COLORS[colors[i % len(colors)]])
        ax_genome.add_patch(patches.Rectangle((start1, h), width=end1-start1, height=0.2, facecolor=mcolors.TABLEAU_COLORS[colors[i % len(colors)]], zorder=1, linewidth=2, clip_on=False))
        ax_genome.add_patch(patches.Rectangle((start2, h), width=end2-start2, height=0.2, facecolor=mcolors.TABLEAU_COLORS[colors[i % len(colors)]], zorder=1, linewidth=2, clip_on=False))
        # plot link
        if row['orient1'] == '+':
            pos1 = (row['end1']-plot_start[row['chrom1']]) + dict_start[row['chrom1']]
        else:
            pos1 = (row['start1']-plot_start[row['chrom1']]) + dict_start[row['chrom1']]
        if row['orient2'] == '+':
            pos2 = (row['start2']-plot_start[row['chrom2']]) + dict_start[row['chrom2']]
        else:
            pos2 = (row['end2']-plot_start[row['chrom2']]) + dict_start[row['chrom2']]
        # height = -0.2
        ax_genome.annotate("",
                    xy=[pos1, h],
                    xytext=[pos2, h],
                    arrowprops=dict(color=mcolors.TABLEAU_COLORS[colors[i % len(colors)]],
                                    arrowstyle="-",
                                    connectionstyle="arc3,rad=-0.3",
                                    linewidth=2
                                    ), zorder=0, clip_on=False)
        h -= 0.3
    ylim = 0 - SV_fragment.shape[0] * 1.5
    ax_genome.set_ylim(ylim, 2)
    
    canvas = FigureCanvas(fig)
    return canvas

def FragmentAssembly(SV_file, resolution, dict_matrix, chrom_list, dict_count, savepath):
    merge = 1000000
    win = 10000000
    # get breakpoint
    bp_result = read_SV_info(SV_file)
    num = bp_result.shape[0]
    SV_fragment = pd.DataFrame(columns=['chrom1', 'start1', 'end1', 'orient1', 'chrom2', 'start2', 'end2', 'orient2'])
    for i in range(num):
        index = [i]
        cur_bp_result = bp_result.loc[index]
        bp_Link, overlap = get_SV_orient(cur_bp_result, dict_matrix, resolution, merge)
        for j in bp_Link.index:
            row = bp_Link.loc[j]
            # smaller that 1Mb region
            if row['pos1'] < (1000000 // resolution) or overlap:
                row_win = 1000000
            else:
                row_win = win
            if row['pos2'] < (1000000 // resolution) or overlap:
                col_win = 1000000
            else:
                col_win = win
            if row['orient1'] == '+':
                pos1_start = row['pos1'] - row_win // resolution
                pos1_end = row['pos1']
            else:
                pos1_start = row['pos1']
                pos1_end = row['pos1'] + row_win // resolution
            if row['orient2'] == '+':
                pos2_start = row['pos2']
                pos2_end = row['pos2'] + col_win // resolution
            else:
                pos2_start = row['pos2'] - col_win // resolution
                pos2_end = row['pos2']
            if pos1_start < 0:
                pos1_start = 0
            if pos2_start < 0:
                pos2_start = 0
            name = row['chrom1'] + '_' + row['chrom2']
            matrix = dict_matrix[name]
            cur_matrix = matrix[pos1_start:pos1_end, pos2_start:pos2_end]
            cut_1, cut_2 = recalculate_breakpoint(cur_matrix, row['orient1'], row['orient2'])
            if row['orient1'] == '-' and row['orient2'] == '-':
                frag1start = pos1_start
                frag1end = pos1_start + cut_1 - 1
                frag2start = pos2_start + cut_2 - 1
                frag2end = pos2_end
            elif row['orient1'] == '+' and row['orient2'] == '+':
                frag1start = pos1_start + cut_1 - 1
                frag1end = pos1_end
                frag2start = pos2_start
                frag2end = pos2_start + cut_2 - 1
            elif row['orient1'] == '+' and row['orient2'] == '-':
                frag1start = pos1_start + cut_1 - 1
                frag1end = pos1_end
                frag2start = pos2_start + cut_2 - 1
                frag2end = pos2_end
            else:
                frag1start = pos1_start
                frag1end = pos1_start + cut_1 - 1
                frag2start = pos2_start
                frag2end = pos2_start + cut_2 - 1
            SV_fragment.loc[len(SV_fragment.index)] = [row['chrom1'], frag1start, frag1end, row['orient1'],
                                                       row['chrom2'], frag2start, frag2end, row['orient2']]
    SV_fragment = coverFragment(SV_fragment)
    all_path, all_orient = constructPath(SV_fragment)
    chrom, start, end, node_list = genomeFragment(SV_fragment)

    dict_start = {}
    dict_end = {}
    for i in range(len(chrom_list)):
        cur_pos = []
        for j in range(len(chrom)):
            if chrom[j] == chrom_list[i]:
                cur_pos.append(start[j])
                cur_pos.append(end[j])
        dict_start[chrom_list[i]] = max(0, min(cur_pos)-5000000//resolution)
        dict_end[chrom_list[i]] = min(max(cur_pos) + 5000000 // resolution, dict_count[chrom_list[i]])

    # plot Hi-C matrix
    chrom_matrix, interval, chrom_text_pos, chrom_text = combine_matrix(dict_matrix, dict_start, dict_end, chrom_list, resolution)
    canvas_matrix = matrix_plot(chrom_matrix, interval, chrom_text_pos, chrom_text, 
                                SV_fragment, chrom, start, end, node_list, chrom_list, dict_start, dict_end)
    
    # plot assembly
    canvas_assembly, CGR_info = plot_SV_genome(all_path, all_orient, SV_fragment)
    return canvas_matrix, canvas_assembly, CGR_info





#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True
import os, sys
path = os.path.split(os.path.abspath(__file__))[0]
sys.path.append(path)
import logging, copy
from collections import defaultdict
import numpy as np
from sklearn.mixture import GaussianMixture
from loadcnv import binCNV
from pomegranate import NormalDistribution, HiddenMarkovModel
logger = logging.getLogger(__name__)


def cbs_stat(x):
    '''
    Given x, Compute the subinterval x[i0:i1] with the maximal segmentation statistic t. 
    Returns t, i0, i1.

    '''
    
    x0 = x - np.mean(x)
    n = len(x0)
    y = np.cumsum(x0)
    e0, e1 = np.argmin(y), np.argmax(y)
    i0, i1 = min(e0, e1), max(e0, e1)
    s0, s1 = y[i0], y[i1]

    return (s1-s0)**2*n/(i1-i0+1)/(n+1-i1+i0), i0, i1+1

def cbs(x, shuffles=1000, p=.05):
    '''
    Given x, find the interval x[i0:i1] with maximal segmentation statistic t. Test that statistic against
    given (shuffles) number of random permutations with significance p.  Return True/False, t, i0, i1; True if
    interval is significant, false otherwise.
    '''
    max_t, max_start, max_end = cbs_stat(x)
    if max_end-max_start == len(x):
        return False, max_t, max_start, max_end
    if max_start < 5:
        max_start = 0
    if len(x)-max_end < 5:
        max_end = len(x)
    thresh_count = 0
    alpha = shuffles*p
    xt = x.copy()
    for i in range(shuffles):
        np.random.shuffle(xt)
        threshold, s0, e0 = cbs_stat(xt)
        if threshold >= max_t:
            thresh_count += 1
        if thresh_count > alpha:
            return False, max_t, max_start, max_end

    return True, max_t, max_start, max_end

def rsegment(x, start, end, L=[], shuffles=1000, p=.05):
    '''
    Recursively segment the interval x[start:end] returning a list
    L of pairs (i,j) where each (i,j) is a significant segment.
    '''
    threshold, t, s, e = cbs(x[start:end], shuffles=shuffles, p=p)
    if (not threshold) | (e-s < 5) | (e-s == end-start):
        L.append((start, end))
    else:
        if s > 0:
            rsegment(x, start, start+s, L)
        if e-s > 0:
            rsegment(x, start+s, start+e, L)
        if start+e < end:
            rsegment(x, start+e, end, L)

    return L

def segment(x, shuffles=1000, p=1e-5):
    '''
    Segment the array x, using significance test based on shuffles rearrangements and significance level p
    '''
    start = 0
    end = len(x)
    L = []
    rsegment(x, start, end, L, shuffles=shuffles, p=p)

    return L

def combine_segment(sig, cbs_seg, hmm_seg, min_seg, min_diff, bufsize=4):

    start = cbs_seg[0][0]
    pool1 = set()
    for b in cbs_seg:
        for i in range(bufsize):
            pool1.add(b[1]-i)
            pool1.add(b[1]+i)

    pool2 = set([b[1] for b in hmm_seg])
    pool = [start] + sorted(pool1 & pool2)
    seg = []
    # combine segments with similar copy numbers
    start, end = pool[0], pool[1]
    if len(pool) > 2:
        for i in range(2, len(pool)):
            ext = pool[i]
            arr1 = sig[start:end]
            mask1 = arr1 == 0
            zero_ratio1 = mask1.sum() / mask1.size
            arr2 = sig[end:ext]
            mask2 = arr2 == 0
            zero_ratio2 = mask2.sum() / mask2.size
            if (mask1.size < min_seg) or (mask2.size < min_seg):
                _add_seg(seg, [start, ext])
                start, end = start, ext
            else:
                if (zero_ratio1 > 0.8) or (zero_ratio2 > 0.8):
                    _add_seg(seg, [start, end])
                    start, end = end, ext
                else:
                    v1 = np.log2(np.median(arr1[arr1!=0]))
                    v2 = np.log2(np.median(arr2[arr2!=0]))
                    if abs(v1 - v2) < min_diff:
                        _add_seg(seg, [start, ext])
                        start, end = start, ext
                    else:
                        _add_seg(seg, [start, end])
                        start, end = end, ext
                        
        if end > seg[-1][1]:
            _add_seg(seg, [start, end])
    else:
        _add_seg(seg, [start, end])

    return seg

def _add_seg(seg, r):

    if not len(seg):
        seg.append(r)
    else:
        if r[0]==seg[-1][0]:
            seg[-1] = r
        else:
            seg.append(r)

class HMMsegment(binCNV):

    def __init__(self, bedgraph, res, nproc=1, ploidy=2, n_states=None):

        binCNV.__init__(self, bedgraph, res)
        self.n_jobs = nproc
        self.ploidy = ploidy
        self.n_states = n_states

    def segment(self, min_seg=5, min_diff=0.4, p=1e-5, max_dist=4):

        outlines = []
        counts = defaultdict(int)
        vMap = {}
        training_seqs, gmm = self.get_states(maxdist=10)
        if self.n_states is None:
            n_states = len(gmm.means_.ravel())
            #logger.info('Estimated HMM state number: {0}'.format(n_states))
            #logger.info('Means of the hidden distributions: {0}'.format(gmm.means_.ravel()))
            #logger.info('Covariances of the hidden distributions: {0}'.format(gmm.covariances_.ravel()))
        else:
            n_states = self.n_states

        if n_states > 1:
            logger.info('Training HMM with {0} states...'.format(n_states))
            model = HiddenMarkovModel.from_samples(
                NormalDistribution,
                n_components=n_states,
                init='kmeans++',
                n_init=10,
                X=training_seqs,
                algorithm='baum-welch',
                stop_threshold=1e-3,
                max_iterations=1000,
                n_jobs=self.n_jobs
            )
            
            logger.info('Done')
        else:
            model = None

        for c in self.bin_cnv:
            logger.info('Segmenting Chromosome {0} ...'.format(c))
            sig = self.bin_cnv[c]
            if not model is None:
                queue = sig[sig > 0]
                seq = np.r_[[s.name for i, s in model.viterbi(np.log2(queue))[1][1:]]]
                hmm_seg = self.assign_cnv(queue, seq)
                predicted = np.zeros(sig.size)
                predicted[sig > 0] = hmm_seg
                hmm_seg = self.call_intervals(predicted)
            else:
                hmm_seg = [(0, sig.size)]

            #logger.info('Combine results from the CBS algorithm ...')
            logsig = np.zeros_like(sig)
            logsig[sig>0] = np.log2(sig[sig>0])
            cbs_seg = segment(logsig, p=p)
            seg = combine_segment(sig, cbs_seg, hmm_seg, min_seg,
                                  min_diff, bufsize=max_dist) # combine results from cbs and hmm
            
            for s, e in seg:
                tmp = sig[s:e]
                mask = tmp==0
                zero_ratio = mask.sum() / mask.size
                if zero_ratio > 0.8:
                    cn = 0
                else:
                    cn = int(np.round(np.median(tmp[tmp!=0])*self.ploidy))
                cur = [c, s, e, cn]
                outlines.append(cur)
                vMap[cn] = cn    
                counts[cn] += (cur[2]-cur[1])
        
        cn_min_count = min(counts.values())
        while cn_min_count < 5: # total number of bins with the a specific copy number must greater than this value
            for i in counts:
                if counts[i]==cn_min_count:
                    cn = i
            neighbor = 0
            mindis = 10000 # a very large number to start
            for i in counts:
                if i == cn:
                    continue
                if np.abs(i - cn) < mindis:
                    neighbor = vMap[i]
                    mindis = np.abs(i - cn)

            tc = counts[cn] + counts[neighbor]

            for i in vMap:
                if vMap[i] == cn:
                    vMap[i] = neighbor
            
            for i in vMap:
                if vMap[i] == neighbor:
                    counts[i] = tc
            
            cn_min_count = min(counts.values())
        
        for line in outlines:
            line[-1] = vMap[line[-1]]
        
        self.outlines = []
        for i, line in enumerate(outlines):
            if not len(self.outlines):
                self.outlines.append(line)
            else:
                if line[0] != self.outlines[-1][0]:
                    self.outlines.append(line)
                else:
                    if (line[3] == 0) and (i < len(outlines)-1) and (line[0] == outlines[i+1][0]) and (line[2] - line[1] < 5):
                        self.outlines[-1][2] = line[2]
                    else:
                        if line[3] != self.outlines[-1][3]:
                            self.outlines.append(line)
                        else:
                            self.outlines[-1][2] = line[2]
    
    def output(self, fil, pre='chr'):

        with open(fil, 'w') as out:
            for c, s, e, v in self.outlines:
                out.write('\t'.join([pre+c, str(s*self.binsize), str(e*self.binsize), str(v)])+'\n')

    
    def pieces(self, arr, minseq=10, scale='log'):

        idx = np.where(arr!=0)[0]
        tmp = np.split(idx, np.where(np.diff(idx)!=1)[0]+1)
        seqs = []
        for i in tmp:
            if i.size >= minseq:
                if scale == 'log':
                    seqs.append(np.log2(arr[i]))
                else:
                    seqs.append(arr[i])
        
        return seqs
    
    def assign_cnv(self, cnv_profile, hmmseq):

        seg = np.zeros(cnv_profile.size)
        for i in np.unique(hmmseq):
            idx = np.where(hmmseq == i)[0]
            tmp = np.split(idx, np.where(np.diff(idx)!=1)[0]+1)
            for t in tmp:
                local_mean = np.median(cnv_profile[t])
                seg[t] = local_mean
        
        return seg
    
    def call_intervals(self, cnvs):

        seg = []
        for i in np.unique(cnvs):
            idx = np.where(cnvs == i)[0]
            tmp = np.split(idx, np.where(np.diff(idx)!=1)[0]+1)
            for t in tmp:
                seg.append((t[0], t[-1]+1))
        
        seg.sort()

        return seg
    
    def get_states(self, maxdist=15):
        """
        Chromosome level estimation.
        """
        training_seqs = []
        whole_genome = np.r_[[]]
        for c in self.bin_cnv:
            arr = self.bin_cnv[c]
            nonzero = arr[arr > 0.1]
            idx = self.hampel_filter(np.log2(nonzero))
            filtered = nonzero[idx]
            whole_genome = np.r_[whole_genome, np.log2(filtered)]

            tmp = np.zeros(nonzero.size)
            tmp[idx] = filtered
            newarr = np.zeros(arr.size)
            newarr[arr > 0.1] = tmp

            self.bin_cnv[c] = newarr
            training_seqs.extend(self.pieces(newarr, scale='log'))
        
        X = whole_genome[:, np.newaxis]
        n_components_range = range(1, maxdist+1)
        _trace_gmm = {}
        aic_arr = []
        for n_components in n_components_range:
            gmm = GaussianMixture(n_components=n_components,
                                  covariance_type='diag',
                                  init_params='k-means++',
                                  n_init=10)
            gmm.fit(X)
            aic = gmm.aic(X)
            aic_arr.append(aic)
            _trace_gmm[n_components] = gmm
        
        aic_arr = np.r_[aic_arr]
        best_idx = aic_arr.argmin()
        gmm = _trace_gmm[best_idx+1]

        return training_seqs, gmm
    
    def hampel_filter(self, arr, ws = 10, n_sigmas = 3):
    
        n = len(arr)
        k = 1.4826 # scale factor for Gaussian distribution
        
        indices = []
        
        for i in range(n):
            if i < ws:
                si = 0
            else:
                si = i - ws
            if i + ws > n:
                ei = n
            else:
                ei = i + ws
            sub = arr[si:ei]
            x0 = np.median(sub)
            S0 = k * np.median(np.abs(sub - x0))
            
            if np.abs(arr[i] - x0) < n_sigmas * S0:
                indices.append(i)
        
        return indices
        


"""
Created on Tue Sep 18 22:15:12 2018

@author: XiaoTao Wang
"""
import numpy as np
import bisect, os
from sklearn.isotonic import IsotonicRegression

def find_chrom_pre(chromlabels):

    ini = chromlabels[0]
    if ini.startswith('chr'):
        return 'chr'
    
    else:
        return ''

def extract_sv_region(hic_pool, c1, c2, pos_1, pos_2, strand_1, strand_2,
                      radius=5000000):
    
    pre = find_chrom_pre(list(hic_pool.chromnames))
    c1 = pre+c1.lstrip('chr')
    c2 = pre+c2.lstrip('chr')

    res = hic_pool.binsize
    if c1==c2:
        if pos_1 > pos_2:
            pos_1, pos_2 = pos_2, pos_1
            strand_1, strand_2 = strand_2, strand_1
    
    p1 = pos_1 // res
    p2 = pos_2 // res
    r = radius // res

    chromsize_1 = hic_pool.chromsizes[c1]
    chromsize_2 = hic_pool.chromsizes[c2]

    if (strand_1=='+') and (strand_2=='+'):
        minx_1 = max(p1-r, 0)
        if c1!=c2:
            minx_2 = max(p2-r, 0)
        else:
            minx_2 = max(p2-r, p1+int((p2-p1)/2)) # handle short inversion ++
        k_p = [minx_1*res, min(p1*res+res,chromsize_1), min(p2*res+res,chromsize_2), minx_2*res]
    elif (strand_1=='+') and (strand_2=='-'):
        minx = max(p1-r, 0)
        maxx = min((p2+r+1)*res, chromsize_2)
        k_p = [minx*res, min(p1*res+res,chromsize_1), p2*res, maxx]
    elif (strand_1=='-') and (strand_2=='-'):
        if c1!=c2:
            maxx_1 = min((p1+r+1)*res, chromsize_1)
        else:
            maxx_1 = min((p1+r+1)*res, (p2-int((p2-p1)/2))*res) # handle short inversion --
        maxx_2 = min((p2+r+1)*res, chromsize_2)
        k_p = [maxx_1, p1*res, p2*res, maxx_2]
    else:
        maxx = min((p1+r+1)*res, chromsize_1)
        minx = max(p2-r, 0)
        k_p = [maxx, p1*res, min(p2*res+res,chromsize_2), minx*res]
    
    return k_p, pos_1, pos_2, strand_1, strand_2


def load_translocation_fil(fil_path, res, minIntra):

    lines = set()
    with open(fil_path, 'r') as source:
        for line in source:
            c1, c2, strands, i1, i2, note = line.rstrip().split()
            if '-' in i1:
                s, e = map(int, i1.split('-'))
                if e - s > res:
                    continue
                p1 = (s+e) // 2
            else:
                p1 = int(i1)
            
            if '-' in i2:
                s, e = map(int, i2.split('-'))
                if e - s > res:
                    continue
                p2 = (s+e) // 2
            else:
                p2 = int(i2)
            s1, s2 = strands[0], strands[1]
            # Filter by intra SV span
            if c1==c2:
                if abs(p2-p1) < minIntra:
                    continue
                if p1 > p2:
                    p1, p2 = p2, p1
                    s1, s2 = s2, s1
                if (s1 == '-') and (s2 == '+') and (abs(p2-p1) < minIntra*2): # duplicate
                    continue
                        
            lines.add((c1.lstrip('chr'), p1, s1, c2.lstrip('chr'), p2, s2, note))
    
    lines = sorted(lines)
    
    return lines

def parse_normal_peaks(peakfile):

    D = {}
    with open(peakfile, 'r') as source:
        for line in source:
            parse = line.rstrip().split()
            if len(parse)==3:
                chrom1 = chrom2 = parse[0].lstrip('chr')
                start1 = int(parse[1])
                start2 = int(parse[2])
            if len(parse)==6:
                chrom1, chrom2 = parse[0].lstrip('chr'), parse[3].lstrip('chr')
                start1 = int(parse[1])
                start2 = int(parse[4])
            if chrom1 > chrom2:
                chrom1, chrom2 = chrom2, chrom1
                start1, start2 = start2, start1
            if chrom1 == chrom2:
                if start1 > start2:
                    start1, start2 = start2, start1

            if (chrom1, chrom2) in D:
                D[(chrom1, chrom2)].append([start1, start2])
            else:
                D[(chrom1, chrom2)] = [[start1, start2]]
    for c in D:
        D[c].sort()
        D[c] = np.r_[D[c]]
    
    return D

def load_sv_peaks(peakfile, maxapart=10000000, onlyInter=True, last='res'):

    byregion = {}
    with open(peakfile, 'r') as source:
        for line in source:
            parse = line.rstrip().split()
            name = parse[0]
            res = int(parse[1])
            c1, p1, c2, p2 = parse[2:6]
            p1, p2 = int(p1), int(p2)
            apart = int(parse[6])
            if parse[-1] == 'False':
                induced = False
            else:
                induced = True
            if apart > maxapart:
                continue
            
            if onlyInter:
                if not induced:
                    continue

            if name in byregion:
                if last=='res':
                    byregion[name].append((c1, p1, c2, p2, int(res)))
                elif last=='full':
                    byregion[name].append((c1, p1, c2, p2, int(res), apart))
            else:
                if last=='res':
                    byregion[name] = [(c1, p1, c2, p2, int(res))]
                else:
                    byregion[name] = [(c1, p1, c2, p2, int(res), apart)]
    
    return byregion

def unique_peaks(byregion, max_mismatch=25000):

    from scipy.spatial import distance_matrix

    # byregion: returned by load_sv_peaks
    visited = {}
    peak_list = []
    for k in byregion:
        for peak in byregion[k]:
            c1, p1, c2, p2 = peak[:4]
            if c1 > c2:
                c1, c2 = c2, c1
                p1, p2 = p2, p1
            if not (c1, c2) in visited:
                visited[(c1, c2)] = np.r_[[[p1, p2]]]
                peak_list.append((c1, p1, c2, p2)+tuple(peak)[4:])
            else:
                dis = distance_matrix(visited[(c1,c2)], [[p1,p2]]).ravel()
                if dis.min() > max_mismatch:
                    visited[(c1, c2)] = np.r_[visited[(c1,c2)], [[p1,p2]]]
                    peak_list.append((c1, p1, c2, p2)+tuple(peak)[4:])
    
    return peak_list


def get_peak_loci(byregion, min_interval=15000):

    bychrom = {}
    for r in byregion:
        for i in byregion[r]:
            if i[4] >= min_interval:
                loci1 = [i[1], i[1]+i[4]]
                loci2 = [i[3], i[3]+i[4]]
            else:
                half = (min_interval - i[4]) // 2
                loci1 = [max(0,i[1]-half), i[1]+i[4]+half]
                loci2 = [max(0,i[3]-half), i[3]+i[4]+half]
            if not i[0] in bychrom:
                bychrom[i[0]] = [loci1]
            else:
                idx = max(0, bisect.bisect(bychrom[i[0]],loci1)-1)
                minp, maxp = loci1
                cache = []
                for q in bychrom[i[0]][idx:]:
                    if q[1] <= loci1[0]:
                        continue
                    if q[0] >= loci1[1]:
                        break
                    if q[0] < minp:
                        minp = q[0]
                    if q[1] > maxp:
                        maxp = q[1]
                    cache.append(q)
                for q in cache:
                    bychrom[i[0]].remove(q)
                bychrom[i[0]].append([minp, maxp])
            
            if not i[2] in bychrom:
                bychrom[i[2]] = [loci2]
            else:
                idx = max(0, bisect.bisect(bychrom[i[2]],loci2)-1)
                minp, maxp = loci2
                cache = []
                for q in bychrom[i[2]][idx:]:
                    if q[1] <= loci2[0]:
                        continue
                    if q[0] >= loci2[1]:
                        break
                    if q[0] < minp:
                        minp = q[0]
                    if q[1] > maxp:
                        maxp = q[1]
                    cache.append(q)
                for q in cache:
                    bychrom[i[2]].remove(q)
                bychrom[i[2]].append([minp, maxp])
            
            bychrom[i[0]].sort()
            bychrom[i[2]].sort()
    
    return bychrom

def set_match(peak_1, peak_2, max_mismatch=40000):

    from collections import defaultdict
    from scipy.spatial import distance_matrix

    only_1 = set()
    only_2 = set()
    overlap_1 = set()
    overlap_2 = set()
    visited = defaultdict(set)
    for r in peak_1:
        parse = r.split(',')
        c1, c2 = parse[1], parse[4]
        if not r in peak_2:
            for p in peak_1[r]:
                only_1.add((r,)+p)
            continue
        ref = [[i[1],i[3]] for i in peak_2[r]]
        query = [[i[1],i[3]] for i in peak_1[r]]
        for v in query:
            dis = distance_matrix([v],ref).ravel()
            argmin = dis.argmin()
            mindis = dis.min()
            if mindis<=min(0.2*abs(v[0]-v[1]), max_mismatch) and (not argmin in visited[r]): # tune the threshold if needed
                overlap_1.add((r,)+(c1,v[0],c2,v[1]))
                visited[r].add(argmin)
            else:
                only_1.add((r,)+(c1,v[0],c2,v[1]))
    
    for r in peak_2:
        for i in range(len(peak_2[r])):
            tmp = (r,) + peak_2[r][i]
            if i in visited[r]:
                overlap_2.add(tmp)
            else:
                only_2.add(tmp)
    
    return only_1, only_2, overlap_1, overlap_2


def peak_match(sv_peaks, normal_peaks, max_mismatch=100000):

    from collections import defaultdict
    from scipy.spatial import distance_matrix

    intra_sv_only = set()
    inter_sv_only = set()
    intra_normal_only = set()
    intra_sv_overlap = set()
    intra_normal_overlap = set()
    visited = defaultdict(set)
    for r in sv_peaks:
        parse = r.split(',')
        c1, c2 = parse[1], parse[4]
        if c1!=c2:
            for peak in sv_peaks[r]:
                inter_sv_only.add((r,)+peak)
            continue
        if not r in normal_peaks:
            for peak in sv_peaks[r]:
                intra_sv_only.add((r,)+peak)
            continue
        ref = [[i[1],i[3]] for i in normal_peaks[r]]
        query = [[i[1],i[3]] for i in sv_peaks[r]]
        for v in query:
            dis = distance_matrix([v],ref).ravel()
            argmin = dis.argmin()
            mindis = dis.min()
            if mindis<=min(0.2*abs(v[0]-v[1]), max_mismatch) and (not argmin in visited[r]): # tune the threshold if needed
                intra_sv_overlap.add((r,)+(c1,v[0],c2,v[1]))
                visited[r].add(argmin)
            else:
                intra_sv_only.add((r,)+(c1,v[0],c2,v[1]))
    
    for r in normal_peaks:
        for i in range(len(normal_peaks[r])):
            tmp = (r,) + normal_peaks[r][i]
            if i in visited[r]:
                intra_normal_overlap.add(tmp)
            else:
                intra_normal_only.add(tmp)          
    
    return intra_sv_only, intra_sv_overlap, intra_normal_only, intra_normal_overlap, inter_sv_only

def map_coordinates(k_p, res, chrom_1, chrom_2, unit='bp'):
    
    # Coordinates on chrom 1
    if k_p[0] < k_p[1]:
        endbin = k_p[1]//res
        if endbin*res < k_p[1]:
            endbin += 1
        startbin = k_p[0]//res
        pl_1 = endbin - startbin
        if unit=='bp':
            D1 = dict(zip(range(k_p[0],k_p[1],res), range(0,pl_1)))
        else:
            D1 = dict(zip(range(startbin,endbin), range(0,pl_1)))
    else:
        endbin = k_p[0]//res
        if endbin*res < k_p[0]:
            endbin += 1
        startbin = k_p[1]//res
        pl_1 = endbin - startbin
        if unit=='bp':
            D1 = dict(zip(range(k_p[1],k_p[0],res), range(pl_1-1,-1,-1)))
        else:
            D1 = dict(zip(range(startbin,endbin), range(pl_1-1,-1,-1)))
    
    # Coordinates on chrom 2
    if k_p[2] < k_p[3]:
        endbin = k_p[3]//res
        if endbin*res < k_p[3]:
            endbin += 1
        startbin = k_p[2]//res
        pl_2 = endbin - startbin
        if unit=='bp':
            D2 = dict(zip(range(k_p[2],k_p[3],res), range(pl_1,pl_1+pl_2)))
        else:
            D2 = dict(zip(range(startbin,endbin), range(pl_1,pl_1+pl_2)))
    else:
        endbin = k_p[2]//res
        if endbin*res < k_p[2]:
            endbin += 1
        startbin = k_p[3]//res
        pl_2 = endbin - startbin
        if unit=='bp':
            D2 = dict(zip(range(k_p[3],k_p[2],res), range(pl_1+pl_2-1,pl_1-1,-1)))
        else:
            D2 = dict(zip(range(startbin,endbin), range(pl_1+pl_2-1,pl_1-1,-1)))
    
    forward = {} # global coordinates to region bin index
    for k in D1:
        forward[(chrom_1, k)] = D1[k]
    for k in D2:
        forward[(chrom_2, k)] = D2[k]
    
    reverse = {v:k for k,v in forward.items()}
    
    return forward, reverse

def properU(pos):
    
    i_part = int(pos) // 1000000 # Integer Part
    d_part = (int(pos) % 1000000) // 1000 # Decimal Part
    
    if (i_part > 0) and (d_part > 0):
        return ''.join([str(i_part), 'M', str(d_part), 'K'])
    elif (i_part == 0):
        return ''.join([str(d_part), 'K'])
    else:
        return ''.join([str(i_part), 'M'])

def _prepare_core(pool_args):
    
    clr, c, maxdis, balance = pool_args

    expected = {} # average over each genomic distance
    hic = clr.matrix(balance=balance, sparse=True).fetch(c)

    if type(balance)==bool:
        weights = clr.bins().fetch(c)['weight'].values
    else:
        weights = clr.bins().fetch(c)[balance].values

    tmp = np.isfinite(weights) & (weights > 0)
    n = hic.shape[0]
    maxdis = min(n-1, maxdis)
    # Assign values for each genomic distance
    for i in range(maxdis+1):
        if i == 0:
            valid = tmp
        else:
            valid = tmp[:-i] * tmp[i:]
        current = hic.diagonal(i)[valid]
        if current.size > 0:
            expected[i] = [current.sum(), current.size]
    
    return expected

def calculate_expected(clr, chroms, maxdis=5000000, balance=True, nproc=1):

    from multiprocess import Pool
    
    res = clr.binsize
    maxdis = maxdis // res
    args = []
    for c in chroms:
        args.append((clr, c, maxdis, balance))
    
    # Allocate processes
    if nproc==1:
        results = list(map(_prepare_core, args))
    else:
        pool = Pool(nproc)
        results = pool.map(_prepare_core, args)
        pool.close()
        pool.join()

    expected = {}
    for i in range(maxdis+1):
        nume = 0
        denom = 0
        for extract in results:
            if i in extract:
                nume += extract[i][0]
                denom += extract[i][1]
        if (denom > 10) and (nume > 0):
            expected[i] = nume / denom
    
    IR = IsotonicRegression(increasing=False, out_of_bounds='clip')
    _d = np.r_[sorted(expected)]
    _y = np.r_[[expected[i] for i in _d]]
    IR.fit(_d, _y)
    d = np.arange(maxdis+1)
    exp = IR.predict(d)
    expected = dict(zip(d, exp))

    return expected

def distance_normaize_core(sub, exp_bychrom, x, y, w):

    # calculate x and y indices
    x_arr = np.arange(x-w, x+w+1).reshape((2*w+1, 1))
    y_arr = np.arange(y-w, y+w+1)

    D = y_arr - x_arr
    D = np.abs(D)
    min_dis = D.min()
    max_dis = D.max()
    if max_dis >= exp_bychrom.size:
        return sub
    else:
        exp_sub = np.zeros(sub.shape)
        for d in range(min_dis, max_dis+1):
            xi, yi = np.where(D==d)
            for i, j in zip(xi, yi):
                exp_sub[i, j] = exp_bychrom[d]
            
        normed = sub / exp_sub

        return normed
    
def image_normalize(arr_2d):

    arr_2d = (arr_2d - arr_2d.min()) / (arr_2d.max() - arr_2d.min()) # value range: [0,1]

    return arr_2d

def distance_normalize(arr_pool, exp_bychrom, xi, yi, w):

    clist = []
    fea = []
    for i in range(xi.size):
        x = xi[i]
        y = yi[i]
        window = arr_pool[i]

        bad_x, bad_y = np.where(np.isnan(window))
        for i_, j_ in zip(bad_x, bad_y):
            window[i_, j_] = 0
             
        if np.count_nonzero(window) < window.size*0.1:
            continue
        
        ll_mean = window[:w, :w].mean()
        if ll_mean > 0:
            center = window[w, w]
            p2LL = center / ll_mean
            if p2LL > 0.1:
                window = distance_normaize_core(window, exp_bychrom, x, y, w)
                fea.append(window)
                clist.append((x, y))
    
    return fea, clist


#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import os, cooler
from cooler import balance
import numpy as np
import pandas as pd
from collections import Counter

def get_marginals(uri, exclude=['M', 'Y', 'MT', 'EBV'], chunksize=int(1e7), nproc=1):

    clr = cooler.Cooler(uri)

    if nproc > 1:
        pool = balance.Pool(nproc)
        map_ = pool.imap_unordered
    else:
        map_ = map

    nnz = clr.info['nnz']
    n_bins = clr.info['nbins']
    edges = np.arange(0, nnz+chunksize, chunksize)
    spans = list(zip(edges[:-1], edges[1:]))

    marg = (
        balance.split(clr, spans=spans, map=map_, use_lock=False)
            .prepare(balance._init)
            .pipe([])
            .pipe(balance._marginalize)
            .reduce(balance.add, np.zeros(n_bins))
    )
    table = clr.bins()[:][['chrom', 'start', 'end']]
    table['Coverage'] = marg.astype(int)
    pool = []
    chroms = [c for c in clr.chromnames if ((not c.lstrip('chr') in exclude) and (not '_' in c))]
    for chrom in chroms:
        pool.append(table[table['chrom']==chrom])
    
    table = pd.concat(pool)

    return table, clr.binsize

# def signal_from_bigwig(table, bw_fil, name='GC'):
#     bw = pyBigWig.open(bw_fil)
#     arr = []
#     for i in table.index:
#         row = table.loc[i]
#         v = bw.stats(row['chrom'], row['start'], row['end'], type='mean')[0]
#         if v is None:
#             v = 0
#         arr.append(v)
#     table[name] = np.r_[arr]
#     return table

def signal_from_txt(table, txt_file, name='GC'):
    txt_data = pd.read_csv(txt_file, sep='\t', compression='gzip')
    txt_data['start'] = txt_data['start'].astype(int)
    txt_data['end'] = txt_data['end'].astype(int)
    chroms = table['chrom'].unique()
    arr = np.zeros(len(table), dtype=float)
    for chrom in chroms:
        chrom_data = txt_data[txt_data['chrom'] == chrom]
        if not chrom_data.empty:
            starts = chrom_data['start'].values
            ends = chrom_data['end'].values
            values = chrom_data['value'].values
            for i, row in table[table['chrom'] == chrom].iterrows():
                overlap_indices = np.where((starts <= row['end']) & (ends >= row['start']))[0]
                mean_value = values[overlap_indices].mean() if overlap_indices.size > 0 else 0
                arr[i] = mean_value
    table[name] = arr
    return table

def count_REsites(table, npz_fil, res):

    RE = np.load(npz_fil)

    RE_by_bin = {}
    for chrom in RE:
        tmp = RE[chrom] // res
        RE_by_bin[chrom] = Counter(tmp)
    
    arr = []
    for i in table.index:
        row = table.loc[i]
        b_i = row['start'] // res
        chrom = row['chrom']
        arr.append(RE_by_bin[chrom][b_i])
    
    table['RE'] = np.r_[arr]

    return table

def filterZeros(table, cols=['GC', 'Mappability', 'RE']):

    mask = (table['GC'] != 0) & (table['Mappability'] != 0) & (table['RE'] != 0) & (table['Coverage'] != 0)
    filtered = table[mask]

    return mask, filtered


    



#!/usr/bin/env python

## Required modules
import os, subprocess, sys, h5py
path = os.path.split(os.path.abspath(__file__))[0]
sys.path.append(path)
from cnv.loadcnv import binCNV
from cnv.correctcnv import matrix_balance
from cnv import runcnv
import cooler
from cnv.segcnv import HMMsegment
import numpy as np
from rpy2.robjects import numpy2ri, Formula
from rpy2.robjects.packages import importr
import wget

def run(hic, genome, enzyme, resolution, nproc, chromosomes, path):
    print(path)
    cachefolder = path + "/result"
    cnv_norm_data_folder = path + "/cnv_norm_data"

    # calculate CNV profile
    weblinks = {
        'hg38_mappability_100mer.1kb.bw': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg38_mappability_100mer.1kb.bw',
        'hg38.MboI.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg38.MboI.npz',
        'hg38.DpnII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg38.MboI.npz',
        'hg38.Arima.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg38.Arima.npz',
        'hg38.BglII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg38.BglII.npz',
        'hg38.uniform.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg38.uniform.npz',
        'hg38.HindIII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg38.HindIII.npz',
        'hg38_mappability_100mer.1kb.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/hg38_mappability_100mer.1kb.txt.gz',
        'hg38_1kb_GC.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/hg38_1kb_GC.txt.gz',
        'hg19.MboI.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg19.MboI.npz',
        'hg19.Arima.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg19.Arima.npz',
        'hg19.HindIII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg19.HindIII.npz',
        'hg19.DpnII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg19.MboI.npz',
        'hg19.BglII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg19.BglII.npz',
        'hg19.uniform.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/hg19.uniform.npz',
        'hg19_mappability_100mer.1kb.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/hg19_mappability_100mer.1kb.txt.gz',
        'hg19_1kb_GC.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/hg19_1kb_GC.txt.gz',
        'mm10.Arima.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm10.Arima.npz',
        'mm10.BglII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm10.BglII.npz',
        'mm10.DpnII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm10.DpnII.npz',
        'mm10.HindIII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm10.HindIII.npz',
        'mm10.MboI.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm10.MboI.npz',
        'mm10.uniform.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm10.uniform.npz',
        'mm10_mappability_100mer.1kb.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/mm10_mappability_100mer.1kb.txt.gz',
        'mm10_1kb_GC.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/mm10_1kb_GC.txt.gz',
        'mm9_mappability_100mer.1kb.bw': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm9_mappability_100mer.1kb.bw',
        'mm9.Arima.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm9.Arima.npz',
        'mm9.BglII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm9.BglII.npz',
        'mm9.DpnII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm9.DpnII.npz',
        'mm9.HindIII.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm9.HindIII.npz',
        'mm9.MboI.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm9.MboI.npz',
        'mm9.uniform.npz': 'http://3dgenome.fsm.northwestern.edu/neoLoopFinder/mm9.uniform.npz',
        'mm9_mappability_100mer.1kb.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/mm9_mappability_100mer.1kb.txt.gz',
        'mm9_1kb_GC.txt.gz': 'https://raw.githubusercontent.com/GaoLabXDU/DeCGR/main/cnv_norm_data/mm9_1kb_GC.txt.gz'
    }
    cachefolder = os.path.abspath(os.path.expanduser(cachefolder))
    if not os.path.exists(cachefolder):
        os.makedirs(cachefolder)
    cnv_file = path + "/result/CNV_profile.txt"
    cnv_seg_file = path + "/result/CNV_seg_profile.txt"
    mapscore_fil = os.path.join(cnv_norm_data_folder, '{0}_mappability_100mer.1kb.txt.gz'.format(genome))
    cutsites_fil = os.path.join(cnv_norm_data_folder, '{0}.{1}.npz'.format(genome, enzyme))
    gc_fil = os.path.join(cnv_norm_data_folder, '{0}_1kb_GC.txt.gz'.format(genome))
    for fil in [mapscore_fil, cutsites_fil, gc_fil]:
        if not os.path.exists(fil):
            key = os.path.split(fil)[1]
            try:
                # Try downloading with wget
                wget.download(weblinks[key], fil)
            except Exception as e:
                return None
            

    mgcv = importr('mgcv')
    stats = importr('stats')
    table, res = runcnv.get_marginals(hic)
    table = runcnv.signal_from_txt(table, gc_fil, name='GC')
    table = runcnv.signal_from_txt(table, mapscore_fil, name='Mappability')
    table = runcnv.count_REsites(table, cutsites_fil, res)
    mask, filtered = runcnv.filterZeros(table)
    fomula = Formula('Coverage ~ s(GC) + s(Mappability) + s(RE)')
    fomula.environment['Coverage'] = numpy2ri.numpy2rpy(filtered['Coverage'].values)
    fomula.environment['GC'] = numpy2ri.numpy2rpy(filtered['GC'].values)
    fomula.environment['Mappability'] = numpy2ri.numpy2rpy(filtered['Mappability'].values)
    fomula.environment['RE'] = numpy2ri.numpy2rpy(filtered['RE'].values)
    gam = mgcv.gam(fomula, family=stats.poisson(link='log'))
    rs = mgcv.residuals_gam(gam, type='working')
    residuals = numpy2ri.rpy2py(rs)
    residuals = residuals - residuals.min()
    idx = np.where(mask)[0]
    CNV = np.zeros(table.shape[0])
    CNV[idx] = residuals
    table['CNV'] = CNV
    bedgraph = table[['chrom', 'start', 'end', 'CNV']]
    bedgraph.to_csv(cnv_file, sep='\t', header=False, index=False)
    print("segmenting cnv")
    # segment CNV profile
    work = HMMsegment(cnv_file,
                      res=resolution,
                      nproc=1,
                      ploidy=2,
                      n_states=None)

    # logger.info('Perform segmentation ...')
    work.segment(
        min_seg=3, min_diff=0.4, max_dist=4, p=1e-5
    )
    work.output(cnv_seg_file)
    print("correcting cnv")
    # correct CNV profile
    clr = cooler.Cooler(hic)
    bins = clr.bins()[:]
    selected_bins = bins[bins['chrom'].isin(chromosomes)].reset_index()
    bin_id_mapping = {old_id: new_id for new_id, old_id in zip(selected_bins.index, selected_bins['index'])}
    pixels = clr.pixels()[:]
    selected_pixels = pixels[pixels['bin1_id'].isin(bin_id_mapping.keys()) & pixels['bin2_id'].isin(bin_id_mapping.keys())]
    selected_pixels.loc[:, 'bin1_id'] = selected_pixels['bin1_id'].map(bin_id_mapping)
    selected_pixels.loc[:, 'bin2_id'] = selected_pixels['bin2_id'].map(bin_id_mapping)
    selected_pixels = selected_pixels.drop_duplicates(subset=['bin1_id', 'bin2_id'])
    new_cool_path = path + "/tmp/test.cool"
    cooler.create_cooler(new_cool_path, selected_bins, selected_pixels)
    hic = new_cool_path
    hic_pool = cooler.Cooler(hic)
    bincnv = binCNV(cnv_seg_file, hic_pool.binsize)
    bincnv.assign_cnv(hic)
    matrix_balance(hic, nproc=nproc, mad_max=5, min_nnz=10, ignore_diags=1)
    return new_cool_path


#! /usr/bin/env python

import argparse
import math
import os
import gzip

# Created by Arya Kaul - 1/12/2017
# Modified by Ferhat Ay - 1/13/2017
# Modified by Nicolas Servant - 1/23/2017
# Modified by Ferhat Ay - 6/5/2017 - added resolution (-r <res>) argument to avoid some problems with inferring it from the first entry of bedFile
# Modified by Ferhat Ay - 6/7/2017 - hitCount to ints, and gzip output 
# Modified by Arya Kaul - 12/21/2017 - bug fix when no -o option specified 
# Modified by Arya Kaul - 05/21/2018 - modified for python3


def outputfithicform(bedPath, matrixPath, intCPath, fragMapPath, biasVectorPath=None, biasVectorOutput=None,res=0):
	print("Loading matrix file...")
	fragDic = {}
	# resolution of data to be determined if res=0 at this point
	with open(bedPath, 'r') as bedFile:
		for lines in bedFile:
			line = lines.rstrip().split()
			chrNum = line[0]
			start = line[1]
			en = line[2]
			if res==0: res=int(en)-int(start) # edge case but if first fragment in file is smaller than res, there is a problem
			mid = int(start)+ int(res/2)
			index = int(line[3])
			fragDic[index] = [chrNum, start, mid, 0] # last field is total contact count: tcc

	lineCount=0
	with open(matrixPath, 'r') as matrixFile:
		with gzip.open(intCPath, 'wt') as interactionCountsFile:
			for lines in matrixFile:
				line = lines.rstrip().split()
				i = int(line[0])
				j = int(line[1])
				cc = float(line[2]) # this can be float or int
				fragDic[i][3] += cc
				fragDic[j][3] += cc
				if cc==int(cc): cc=int(cc) # make sure to convert to integer if it ends with ".0"
				interactionCountsFile.write(str(fragDic[i][0])+'\t'+str(fragDic[i][2])+'\t'+\
					str(fragDic[j][0])+'\t'+str(fragDic[j][2])+'\t'+str(cc)+"\n")
				lineCount+=1
				if lineCount%1000000==0: print ("%d million lines read" % int(lineCount/1000000))

	with gzip.open(fragMapPath, 'wt') as fragmentMappabilityFile:
		for indices in sorted(fragDic): # sorted so that we start with the smallest index
			toWrite = 0
			if fragDic[indices][3]> 0: toWrite=1 
			fragmentMappabilityFile.write(str(fragDic[indices][0])+'\t'+str(fragDic[indices][1])+'\t'+\
				str(fragDic[indices][2])+'\t'+str(int(fragDic[indices][3]))+'\t'+str(toWrite)+'\n')

	if biasVectorPath is not None and biasVectorOutput is not None:
		print ("Converting bias file...")

		biasVec=[0,0] # bias sum and biasCount
		biasDic={} # bias for each index
		i=1 # one-based indices
		with open(biasVectorPath, 'r') as biasVectorFile:
				for lines in biasVectorFile:
					value = float(lines.rstrip()) #just one entry that can be nan or a float
					index = int(i)
					i+=1
					biasDic[index]=value
					if not math.isnan(value):
						biasVec[0]+=value #sum
						biasVec[1]+=1 # count

		# Centering the bias values on 1.
		biasAvg=biasVec[0]/biasVec[1]

		with gzip.open(biasVectorOutput, 'wt') as biasVectorOutputFile:
			for index in sorted(biasDic):
				value=biasDic[index]
				if not math.isnan(value):
					value=value/biasAvg
				else: 
					value=-1
				biasVectorOutputFile.write(str(fragDic[index][0])+'\t'+str(fragDic[index][2])+'\t'+str(value)+'\n')
	print("Conversion from HiC-Pro to Fit-Hi-C format completed")

#outputfithicform(args.bedPath, args.matrixPath, args.intCPath, args.fragMapPath, args.biasVectorPathandOutput[0], args.biasVectorPathandOutput[1])

def main():
	# Example without bias files
	outputfithicform('raw/1000000/hIMR90_HindIII_r1_1000000_abs.bed', 'raw/1000000/hIMR90_HindIII_r1_1000000.matrix', 'fithic.interactionCounts', 'fithic.fragmentMappability')
	# Example with bias files
	outputfithicform('raw/1000000/hIMR90_HindIII_r1_1000000_abs.bed', 'raw/1000000/hIMR90_HindIII_r1_1000000.matrix', 'fithic.interactionCounts', 'fithic.fragmentMappability','hicpro.biases','fithic.biases')

if __name__=="__main__":

	parser = argparse.ArgumentParser()
	parser.add_argument("-i", "--matrix", help="Input matrix file with raw contact frequencies.", required=True)
	parser.add_argument("-b", "--bed", help="BED file with bins coordinates.", required=True)
	parser.add_argument("-s", "--bias", help="The bias file provided after IC normalization.", default=None)
	parser.add_argument("-o", "--output", help="Output path", default=".")
	parser.add_argument("-r", "--resolution", help="Resolution of the matrix", type=int, default=0) # 0 means it is inferred from fragments file's first entry

	args = parser.parse_args()

	icounts_output = os.path.join(args.output + "/fithic.interactionCounts.gz")
	fragmap_output = os.path.join(args.output + "/fithic.fragmentMappability.gz")
	bias_output = None

	if args.bias is not None:
		bias_output = os.path.join(args.output + "/fithic.biases.gz")

	outputfithicform(args.bed, args.matrix, icounts_output, fragmap_output, args.bias, bias_output, args.resolution)




##############################################################################
### To use the functions in this lib simply import this python module using
### import myUtils
### Then you'll able able to call functions with the proper arguments using
### returnVal=myUtils.func1(arg1,arg2)
##############################################################################
##############################################################################
import numpy as np
import math

################### FUNC scale_a_list  #######################################
#### Given a list this functions multiplies it with the given scaling parameter.
##############################################################################
def scale_a_list(somelist, s):
    return [1.0*i*s for i in somelist]

################### FUNC chr_name_conversion  #####################################
#### Given an identifier for the chromosome name (str) or a chromosome number (int) 
###  and an organism this function converts the identifier to the other representation. 
### Example: 

################### FUNC sort_by_column  #####################################
#### Given a list with 1 or more columns this functions sorts it according to 
### the desired column n [0 len(list)). Does this in-place.
##############################################################################
def sort_by_column(somelist, n):
    somelist[:] = [(x[n], x) for x in somelist]
    somelist.sort()
    somelist[:] = [val for (key, val) in somelist]
    return

################### FUNC chr_name_conversion  #####################################
#### Given an identifier for the chromosome name (str) or a chromosome number (int) 
###  and an organism this function converts the identifier to the other representation. 
### Example: 
### converts 'chrX' or 'X' to 23 for human
### converts 23 to 'chrX' 1 to 'chr1' for human
##############################################################################
def chr_name_conversion(chrIn,org):
    if isinstance(chrIn, int): # int to str
        if org=='human':
            if  chrIn<23 and chrIn>0:
                chrOut='chr'+str(chrIn)
            elif chrIn==23:
                chrOut='chrX'
            elif chrIn==24:
                chrOut='chrY'
            else:
                return 'problem'
        elif org=='mouse':
            if  chrIn<20 and chrIn>0:
                chrOut='chr'+str(chrIn)
            elif chrIn==20:
                chrOut='chrX'
            elif chrIn==21:
                chrOut='chrY'
            else:
                return 'problem'
        else:
            chrOut='chr'+str(chrIn)
    else: # str to int
        if 'chr' in chrIn:
            chrIn=chrIn[:3] # cut the 'chr'
        if org=='human':
            if  chrIn=='X':
                chrOut=23
            elif chrIn=='Y':
                chrOut=24
            else:
                chrOut=int(chrIn)
        elif org=='mouse':
            if  chrIn=='X':
                chrOut=20
            elif chrIn=='Y':
                chrOut=21
            else:
                chrOut=int(chrIn)

    return chrOut

################### FUNC in_range_check  #####################################
####  Check whether the given interactionDistance is within the range we are 
###   interested. Should only be used for intra-chromosomal interactions.
##############################################################################
def in_range_check(interactionDistance,distLowThres,distUpThres):
    # modification - sourya
    # previously interactionDistance > distLowThres
    # now interactionDistance >= distLowThres
    if (distLowThres==-1 or (distLowThres>-1 and interactionDistance >= distLowThres)) and\
        (distUpThres==-1 or (distUpThres>-1 and interactionDistance <= distUpThres)):
        return True
    return False

################### CLASS Interaction ########################################
####  This class is a container for interactions between two loci as observed 
###   in Hi-C datasets. 
##############################################################################
class Interaction:
    chr1='chr'
    mid1=-1
    chr2='chr'
    mid2=-1
    hitCount=0
    distance=-1
    type='null'
    pval=-1.0
    qval=-1.0
    dictkey='null'
    def __init__(self):
        self.data = []

    def __init__(self, locusPair):
        self.chr1=locusPair[0]
        self.mid1=int(locusPair[1])
        self.chr2=locusPair[2]
        self.mid2=int(locusPair[3])
        if self.chr1==self.chr2:
            self.type='intra'
            self.distance=abs(self.mid1-self.mid2)
        else:
            self.type='inter'

    def setCount(self,x):
        self.hitCount=int(x)
    def setType(self,x):
        self.type=str(x)
    def setPval(self,x):
        self.pval=float(x)
    def setQval(self,x):
        self.qval=float(x)
    def getDistance(self):
        return self.distance
    def getCount(self):
        return self.hitCount
    def getType(self,distLowThres,distUpThres):
        if self.type == 'inter':
            return self.type
        # modification - sourya
        # previously self.distance > distLowThres
        # now self.distance >= distLowThres
        if (distLowThres==-1 or (distLowThres>-1 and self.distance >= distLowThres)) and\
            (distUpThres==-1 or (distUpThres>-1 and self.distance <= distUpThres)):
            self.type='intraInRange'
        elif (distLowThres>-1 and self.distance <= distLowThres):
            self.type='intraShort'
        elif (distUpThres>-1 and self.distance > distUpThres):
            self.type='intraLong'
        return self.type

        #if self.chr1==min(self.chr1,self.chr2):
                
################### CLASS Locus ##############################################
####  This class is a container for loci in general and for Hi-C data.
##############################################################################
class Locus:
    chr='chr'
    mid=-1
    #start=-1
    #end=-1
    type='null'
    hitCount=-1
    #avgGCcontent=-1.0
    #avgMappability=-1.0
    def __init__(self):
        self.data = []

    def __init__(self, locus):
        self.chr=locus[0]
        self.mid=int(locus[1])
    def setCount(self,x):
        self.hitCount=int(x)
    def setType(self,x):
        self.type=str(x)


################### FUNC convert_UCSC_to_bed_format  #####################################
#### Given a locus in UCSC format this function converts it to bed format with 3 fields
### chr1:121-21111 --> ['chr1', 121, 21111]
##############################################################################
def convert_UCSC_to_bed_format(l):
    chr=l[:l.find(':')]
    st=int(l[l.find(':')+1:l.find('-')])
    en=int(l[l.find('-')+1:])
    return (chr,st,en)


################### FUNC read_bed_for_chrlist  #####################################
#### Reads a bed file and extracts a dictionary for each chromosome given in the chromosomeList
### <noOfAdditionalFields> dditional fields are going to be considered on top of first 3 fields.
##############################################################################
def read_bed_for_chrlist(bedFile,chromosomeList,noOfAdditionalFields):
    dic={}
    for ch in chromosomeList:
        dic[ch]=[]

    infile=open(bedFile,'r')
    for line in infile:
        ## line is too short to be real
        if len(line)<3:
            break
        words=line.rstrip().split()
        thisChr=words[0]
        if thisChr not in chromosomeList:
            continue
        thisStart=int(words[1])
        thisEnd=int(words[2])
        additions=[]
        if noOfAdditionalFields>0:
            for n in range(noOfAdditionalFields):
                additions.append(words[3+n])
        dic[thisChr].append([thisStart, thisEnd]+additions)
    infile.close()
    for ch in chromosomeList:
        dic[ch]=sorted(dic[ch])
    return dic

################### FUNC get_overlap_between_intervals  #####################################
#### Finds the overlap between two intervals end points inclusive
### a=[10,20]; b=[20,30] --> f(a,b)=1
### a=[10,20]; b=[15,30] --> f(a,b)=6
##############################################################################
def get_overlap_between_intervals(a, b):
    return max(0, min(a[1], b[1]) - max(a[0], b[0])+1)






"""
Setup script for DeCGR.

"""
import os, sys, glob
import setuptools

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if (sys.version_info.major!=3) or (sys.version_info.minor<7):
    print('PYTHON 3.7+ IS REQUIRED. YOU ARE CURRENTLY USING PYTHON {}'.format(sys.version.split()[0]))
    sys.exit(2)

# Guarantee Unix Format
for src in glob.glob('scripts/*'):
    text = open(src, 'r').read().replace('\r\n', '\n')
    open(src, 'w').write(text)

setuptools.setup(
    name = 'DeCGR',
    version = "1.1.10",
    author = "Li Junping",
    author_email = 'lijunping02@qq.com',
    url = 'https://github.com/GaoLabXDU/DeCGR',
    description = 'An interactive toolkit for deciphering CGRs from Hi-C data',
    keywords = ("Hi-C", "complex rearrangements", "3D genome"),
    scripts = glob.glob('scripts/*'),
    packages = setuptools.find_packages(),
    include_package_data = True,
   
    platforms = "any",
    license="MIT Licence",
    install_requires = [
        "pandas",
        "scikit-learn",
        "PyQt5==5.15.11",
        "numpy",
        "wget",
        "cooler",
        "pomegranate==0.15.0",
        "matplotlib",
        "sortedcontainers",
        "scikit-image"
        ]
    )



#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import logging, warnings, h5py, multiprocess
import numpy as np
from cooler import balance, util, Cooler
from collections import defaultdict

# override the original functions

logger = logging.getLogger(__name__)

def _balance_genomewide(bias, cnv, ucnv, clr, spans, filters, map, tol, max_iters,
                        rescale_marginals, use_lock):
    scale = 1.0
    n_bins = len(bias)

    for _ in range(max_iters):
        marg = (
            balance.split(clr, spans=spans, map=map, use_lock=use_lock)
                .prepare(balance._init)
                .pipe(filters)
                .pipe(balance._timesouterproduct, bias)
                .pipe(balance._marginalize)
                .reduce(balance.add, np.zeros(n_bins))
        )
        
        nzmarg = marg[marg != 0]
        if not len(nzmarg):
            scale = np.nan
            bias[:] = np.nan
            var = 0.0
            break
        
        var_pool = []
        weights = []
        for uc in ucnv:
            if uc==0:
                continue
            umarg = marg[cnv==uc]
            tmp = umarg[umarg != 0]
            if not len(tmp):
                bias[cnv==uc] = 0
            else:
                umarg = umarg / tmp.mean()
                umarg[umarg==0] = 1
                bias[cnv==uc] = bias[cnv==uc] / umarg
                var_pool.append(tmp.var())
                weights.append(tmp.size)

        var = np.average(var_pool, weights=weights)
        logger.info("variance is {}".format(var))
        if var < tol:
            break
    else:
        warnings.warn(
            'Iteration limit reached without convergence.',
            balance.ConvergenceWarning)
    
    bias[bias==0] = np.nan
    for uc in ucnv:
        if uc==0:
            continue
        umarg = marg[cnv==uc]
        tmp = umarg[umarg != 0]
        if len(tmp):
            scale = tmp.mean()
            if rescale_marginals:
                bias[cnv==uc] = bias[cnv==uc] / np.sqrt(scale)

    '''
    # might be the reason why this normalized matrix has different range from ICE
    scale = nzmarg.mean()
    bias[bias == 0] = np.nan
    if rescale_marginals:
        bias /= np.sqrt(scale)
    '''
    return bias, var

def iterative_correction(clr, chunksize=int(1e7), map=map, tol=1e-5, min_nnz=10,
                         min_count=0, mad_max=5, ignore_diags=1, max_iters=200,
                         rescale_marginals=True, use_lock=False):
    
    # CNV must exists
    cnv = clr.bins()['CNV'][:].values
    # Divide the number of elements into non-overlapping chunks
    nnz = clr.info['nnz']
    if chunksize is None:
        chunksize = nnz
        spans = [(0, nnz)]
    else:
        edges = np.arange(0, nnz+chunksize, chunksize)
        spans = list(zip(edges[:-1], edges[1:]))

    # List of pre-marginalization data transformations
    base_filters = []
    if ignore_diags:
        base_filters.append(balance.partial(balance._zero_diags, ignore_diags))

    # Initialize the bias weights
    n_bins = clr.info['nbins']
    bias = np.ones(n_bins, dtype=float)

    # Drop bins with too few nonzeros from bias
    if min_nnz > 0:
        filters = [balance._binarize] + base_filters
        marg_nnz = (
            balance.split(clr, spans=spans, map=map, use_lock=use_lock)
                .prepare(balance._init)
                .pipe(filters)
                .pipe(balance._marginalize)
                .reduce(balance.add, np.zeros(n_bins))
        )
        bias[marg_nnz < min_nnz] = 0

    filters = base_filters
    marg = (
        balance.split(clr, spans=spans, map=map, use_lock=use_lock)
            .prepare(balance._init)
            .pipe(filters)
            .pipe(balance._marginalize)
            .reduce(balance.add, np.zeros(n_bins))
    )

    # Drop bins with too few total counts from bias
    if min_count:
        bias[marg < min_count] = 0

    # MAD-max filter on the marginals
    if mad_max > 0:
        offsets = clr._load_dset('indexes/chrom_offset')
        for lo, hi in zip(offsets[:-1], offsets[1:]):
            c_marg = marg[lo:hi]
            marg[lo:hi] /= np.median(c_marg[c_marg > 0])
        logNzMarg = np.log(marg[marg>0])
        med_logNzMarg = np.median(logNzMarg)
        dev_logNzMarg = balance.mad(logNzMarg)
        cutoff = np.exp(med_logNzMarg - mad_max * dev_logNzMarg)
        bias[marg < cutoff] = 0
    
    bias[cnv==0] = 0
    ucnv = np.unique(cnv)
    # Do balancing
    bias, var = _balance_genomewide(
            bias, cnv, ucnv, clr, spans, base_filters, map, tol, max_iters,
            rescale_marginals, use_lock)

    stats = {
        'tol': tol,
        'min_nnz': min_nnz,
        'min_count': min_count,
        'mad_max': mad_max,
        'ignore_diags': ignore_diags,
        'converged': var < tol,
        'var': var,
    }

    return bias, stats

def matrix_balance(cool_uri, nproc=1, chunksize=int(1e7), mad_max=5,
                min_nnz=10, min_count=0, ignore_diags=1, tol=1e-5,
                max_iters=1000):
    '''
    Perform separate matrix balancing for regions with different copy numbers
    and output the bias vector in the "sweight" column.
    '''
    cool_path, group_path = util.parse_cooler_uri(cool_uri)
    # Overwrite the existing sweight column
    with h5py.File(cool_path, 'r+') as h5:
        grp = h5[group_path]
        if 'sweight' in grp['bins']:
            del grp['bins']['sweight']
    
    clr = Cooler(cool_uri)
    
    try:
        if nproc > 1:
            pool = multiprocess.Pool(nproc)
            map_ = pool.imap_unordered
        else:
            map_ = map
        
        bias, stats = iterative_correction(
                clr,
                chunksize=chunksize,
                tol=tol,
                min_nnz=min_nnz,
                min_count=min_count,
                mad_max=mad_max,
                max_iters=max_iters,
                ignore_diags=ignore_diags,
                rescale_marginals=True,
                use_lock=False,
                map=map_)
    finally:
        if nproc > 1:
            pool.close()
    
    if not stats['converged']:
        logger.error('Iteration limit reached without convergence')
        logger.error('Storing final result. Check log to assess convergence.')
    
    with h5py.File(cool_path, 'r+') as h5:
        grp = h5[group_path]
        # add the bias column to the file
        h5opts = dict(compression='gzip', compression_opts=6)
        grp['bins'].create_dataset('sweight', data=bias, **h5opts)
        grp['bins']['sweight'].attrs.update(stats)
